#include <xc.h>

#define uint8   unsigned char   // 0-255
#define uint16  unsigned short  // 0-65432

/*
 * Rx data is 9600/8/N/1. 
 * Expecting 7 ASCII chars + CR (0x13) Rn=0000<cr> Where n=0-7 and 0000 is the hex value to display in binary
 * If no data received within 10 secs then will begin random stuff 
 */

// Press the help icon left of the dashboard and select configuration registers for pragma/config settings

#pragma config FOSC     = ECH     // Use external 11.0592Mhz OSC
#pragma config PLLEN    = OFF     // No PLL
#pragma config WDTE     = OFF     // Watchdog timer OFF
#pragma config BOREN    = OFF      // Turn on Brown-out reset so the device resets on any power problem
#pragma config BORV     = HI      // Brown out reset HI
#pragma config STVREN   = OFF      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config CLKOUTEN = OFF     // Clk out as IO

#define RESET_PIN(state)    LATAbits.LATA1 = state // Output - Shift Reg 74HC595 Reset (Active low)
#define LAT_PIN(state)      LATAbits.LATA3 = state // Output - Shift Reg 74HC595 output D-Latch clock
#define SERDAT_PIN(state)   LATAbits.LATA0 = state // Output - Shift Reg 74HC595 Serial In Data
#define SERCLK_PIN(state)   LATAbits.LATA2 = state // Output - Shift Reg 74HC595 Master In Data Clock
#define C1(state)           LATCbits.LATC4 = state // Output - Col Driver
#define C2(state)           LATCbits.LATC0 = state // Output - Col Driver
#define C3(state)           LATCbits.LATC5 = state // Output - Col Driver
#define C4(state)           LATCbits.LATC1 = state // Output - Col Driver
#define C5(state)           LATAbits.LATA5 = state // Output - Col Driver
#define C6(state)           LATCbits.LATC3 = state // Output - Col Driver
#define C7(state)           LATAbits.LATA6 = state // Output - Col Driver
#define C8(state)           LATCbits.LATC2 = state // Output - Col Driver

#define VoltageSettleDelay  NOP(); NOP(); NOP();

// Port B RB0-RB3 used as address control dip switch inputs rest N/U
// Port C7 UART RXD Input

volatile unsigned short LEDRegister[8], mask, res, res2;

volatile unsigned char  HostActiveTimeout=0;
volatile unsigned char  RxBuffer[21];
volatile unsigned char  RxBufPtr=0;
volatile unsigned char  RxDataRdy=0;
volatile unsigned char  addr=0;
volatile unsigned char  i, ColSel, RegisterSelect;

void USdelay(unsigned char a)
{
    while(a--) NOP();
}

void ClockData(void)
{
    SERCLK_PIN(0);
    NOP(); NOP(); NOP(); NOP(); 
    SERCLK_PIN(1);
    NOP(); NOP(); NOP(); NOP(); 
    SERCLK_PIN(0);
}

void ClockLatch(void)
{
    LAT_PIN(0);
    NOP(); NOP(); NOP(); NOP(); 
    LAT_PIN(1);
    NOP(); NOP(); NOP(); NOP(); 
    LAT_PIN(0);
}

void __interrupt() TimerAndUartInterrupt()
{    
    if(PIR1bits.RCIF)
    {
        if (RCSTAbits.FERR || RCSTAbits.OERR) // reset any errors
        {
          RCSTAbits.CREN = 0;
          RCSTAbits.CREN = 1;
        }  
        //read the byte from receive register into receive buffer
        if(RxBufPtr<20)
        {  
          RxBuffer[RxBufPtr++] = RCREG;
          if(RCREG==0x0d) RxDataRdy=1; // message received so decode
        }
        PIR1bits.RCIF=0; // re-enable interrupt for next char
    }   
    if(INTCONbits.T0IF) // clock Selected register into the 74HC595 shift register. ensure takes < 740uS !!!
    {
        mask=0x8000;
        while(mask)
        {                
            res = LEDRegister[RegisterSelect] & mask;
            if(res==0) SERDAT_PIN(1); else SERDAT_PIN(0);
            ClockData();
            mask=mask>>1;
        }    
        ClockLatch();

        // enable the correct scan column we are displaying (Selected reg LED's ON) and disable all other cols so other LEDs OFF
        if(ColSel==0) C1(1); else C1(0);
        if(ColSel==1) C2(1); else C2(0);
        if(ColSel==2) C3(1); else C3(0);
        if(ColSel==3) C4(1); else C4(0);
        if(ColSel==4) C5(1); else C5(0);
        if(ColSel==5) C6(1); else C6(0);
        if(ColSel==6) C7(1); else C7(0);
        if(ColSel==7) C8(1); else C8(0);

        ColSel++;
        if(ColSel > 7) ColSel=0; // Column select goes 0-7 (Note can make this bigger than 8 to DIM LED brightness)

        RegisterSelect++;
        if(RegisterSelect > 7) RegisterSelect=0; // Register select goes 0-7 (Keep separate from ColSel in case we want brightness altered)
        INTCONbits.T0IF=0;
    }    
}

unsigned char ascii2dec( unsigned char data) // expecting a hexedecimal num 0..F anything else returns 0
{
    if(data=='0') return 0;
    if(data=='1') return 1;
    if(data=='2') return 2;
    if(data=='3') return 3;
    if(data=='4') return 4;
    if(data=='5') return 5;
    if(data=='6') return 6;
    if(data=='7') return 7;
    if(data=='8') return 8;
    if(data=='9') return 9;
    if(data=='A') return 10;
    if(data=='B') return 11;
    if(data=='C') return 12;
    if(data=='D') return 13;
    if(data=='E') return 14;
    if(data=='F') return 15;
    if(data=='a') return 10;
    if(data=='b') return 11;
    if(data=='c') return 12;
    if(data=='d') return 13;
    if(data=='e') return 14;
    if(data=='f') return 15;
    return 0; // default/error
}

int main()
{
    TRISA   = 0x00; // PortA: All Output
    ANSELA  = 0x00;
    LATA    = 0x00;
    TRISB   = 0x0f; // PortB: RB0-Rb3 input / rest output
    ANSELB  = 0x00;
    LATB    = 0x00;
    WPUB    = 0x0f; // pull up address bits pulled down if sw set
    TRISC   = 0x80; // PortC: All output except RC7 for RXD
    LATC    = 0x00;

    addr = (~PORTB) & 0x0f; // dip switch returns 0-15 (HAVE PULL UP SO ON=0)
    RxDataRdy = 0;
    RegisterSelect = 0;
    ColSel = 0;
    HostActiveTimeout = 0;  
    for(i=0; i<8; i++) LEDRegister[i] = 0XC0F3>>addr;
    for(i=0; i<20; i++) RxBuffer[i]=0;
    
    RESET_PIN(1); // remove 74HC595D reset to allow data shifting
    ClockLatch();
    USdelay(200);

    // configure serial UART (connected to USB)  9600 baud 8/N/1 (Calculations based on OSC=11.0952Mhz P325 datasheet)
    TXSTAbits.BRGH = 0;
    TXSTAbits.SYNC= 0;  // Async
    BAUDCONbits.BRG16 = 0;
    SPBRG = 17;
    RC1STAbits.SPEN=1;  // Enable Serial
    TXSTAbits.TXEN= 0;  // TX DISABLED
    RCSTAbits.CREN= 1;  // RX ENABLED
    PIE1bits.RCIE = 1;  // enable receive interrupt
    INTCONbits.PEIE=1;  // enable interrupts for UART  

    // configure timer 0 interrupt: tick = Fosc (11.0592Mhz) /4/8/256 = approx every 740uS
    OPTION_REG = 0x02;
    INTCONbits.T0IE=1;

    // configure timer 1 (as a 16 bit counter, tick = Fosc/4/8 = 2.88uS
    T1CON  = 0x31;
    T1GCON = 0;
    TMR1L  = 0;
    TMR1H  = addr*16; // so multiple panels change at different times if all powered up together  

    INTCONbits.GIE =1;  // enable interrupts
    
    while(1)
    {
        CLRWDT();

        if(TMR1H>250) // approx every 188mS update the serial data timeout and if no data received after 5 seconds then give a rand pattern
        {  
            TMR1H=0;
            if(HostActiveTimeout<250)
                HostActiveTimeout++;
            else
            {
                // no data coming in so do something interesting
                for(i=0;i<8;i++)                    
                  LEDRegister[i]=LEDRegister[i]+addr+i;
            }    
        }
        if(RxDataRdy) // RXD decode 9 char message received through serial and set regLED[n] (expecting format AnRn=0000) Where A=Panel Address, R=Register and n = 0..F eg "A0R4=1234"
        {
            if(RxBuffer[0]=='A' && ascii2dec(RxBuffer[1])==addr && RxBuffer[2]=='R' && RxBuffer[4]=='=') // format still good
            {
                i=ascii2dec(RxBuffer[3])-addr; // decode the reg number ASCII to dec
                if (i>=0 && i<8)
                {    
                    res2 = (ascii2dec(RxBuffer[5])*4096) + (ascii2dec(RxBuffer[6])*256) + (ascii2dec(RxBuffer[7])*16) + ascii2dec(RxBuffer[8]);
                    LEDRegister[i]=res2;
                }    
                HostActiveTimeout=0; // valid message, reset timeout
            }   
            RxDataRdy=0;
            RxBufPtr=0;
        }  
    }  
}
